import { RealmDefinition } from "./RealmDefinition";
import * as Realm from 'realm';
export declare class UserMetadataRow {
    static schema: Realm.ObjectSchema;
    user: User;
    key: string;
    value: string;
    toJSON(): {
        key: string;
        value: string;
    };
}
export declare class User {
    static schema: {
        name: string;
        primaryKey: string;
        properties: {
            userId: {
                type: string;
                optional: boolean;
            };
            isAdmin: {
                type: string;
                optional: boolean;
            };
            accounts: {
                type: string;
                objectType: string;
            };
            metadata: {
                type: string;
                objectType: string;
                default: any[];
                optional: boolean;
            };
        };
    };
    userId: string;
    isAdmin: boolean;
    accounts?: Account[];
    metadata?: UserMetadataRow[];
    created?: boolean;
    toJSON(): {
        user_id: string;
        is_admin: boolean;
        accounts: {
            provider: string;
            provider_id: string;
        }[];
        metadata: {
            key: string;
            value: string;
        }[];
    };
}
export declare class Account {
    static schema: Realm.ObjectSchema;
    provider: string;
    providerId: string;
    user?: User;
    toJSON(): {
        provider: string;
        provider_id: string;
    };
}
export declare class Permission {
    static schema: Realm.ObjectSchema;
    user: User;
    realmFile: RealmFile;
    mayRead: boolean;
    mayWrite: boolean;
    mayManage: boolean;
    updatedAt: Date;
}
export declare class RealmFile {
    static schema: Realm.ObjectSchema;
    path: string;
    owner?: User | null;
    syncLabel: string;
    createdAt: Date;
    permissions?: Realm.List<Permission>;
}
export declare const AdminRealm: RealmDefinition;
