"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __param = (this && this.__param) || function (paramIndex, decorator) {
    return function (target, key) { decorator(target, key, paramIndex); }
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const decorators_1 = require("../decorators");
const Server_1 = require("../Server");
const Token_1 = require("../shared/Token");
const realms_1 = require("../realms");
const errors = require("../errors");
const path = require("path");
const moment = require("moment");
const URI = require("urijs");
const _ = require("lodash");
const events_1 = require("events");
let PermissionService = PermissionService_1 = class PermissionService extends events_1.EventEmitter {
    constructor() {
        super(...arguments);
        this.handleAdminRealmPermissionChange = (collection, change) => {
            this.emit('handlePermissionChange', collection);
            const insertedPermissions = change.insertions.map(i => collection[i]);
            const modifiedPermissions = change.modifications.map(i => collection[i]);
            const deletedPermissions = change.modifications.map(i => collection[i]);
            const insertedAndChangedPermission = insertedPermissions.concat(modifiedPermissions);
            for (let permission of insertedAndChangedPermission) {
                const { userSpecificRealm, userSpecificPermission } = this.getUserSpecificPermissionAndAssociatedRealm(permission.realmFile.path, permission.user.userId);
                userSpecificRealm.write(() => {
                    if (!userSpecificPermission) {
                        userSpecificRealm.create('Permission', {
                            mayManage: permission.mayManage,
                            mayWrite: permission.mayWrite,
                            mayRead: permission.mayManage,
                            path: permission.realmFile.path,
                            userId: permission.user.userId,
                            updatedAt: new Date()
                        });
                    }
                    else {
                        userSpecificPermission.mayManage = permission.mayManage;
                        userSpecificPermission.mayRead = permission.mayRead;
                        userSpecificPermission.mayWrite = permission.mayWrite;
                        userSpecificPermission.path = permission.realmFile.path;
                        userSpecificPermission.updatedAt = new Date();
                        userSpecificPermission.userId = permission.user.userId;
                    }
                });
            }
            for (let permission of deletedPermissions) {
                const { userSpecificRealm, userSpecificPermission } = this.getUserSpecificPermissionAndAssociatedRealm(permission.realmFile.path, permission.user.userId);
                if (!userSpecificPermission) {
                }
                else {
                    userSpecificRealm.write(() => {
                        userSpecificRealm.delete(userSpecificPermission);
                    });
                }
            }
        };
    }
    get adminRealm() {
        return this._syncedAdminRealm.realm;
    }
    get starPermissionsRealm() {
        return this._syncedStarPermissionsRealm.realm;
    }
    serverStarted(server) {
        this.server = server;
        this.dataPath = path.resolve(server.config.dataPath, "permissions");
        this.privateKey = server.privateKey;
        this.publicKey = server.publicKey;
        this.logger = server.logger;
        this.adminTokenUser = Realm.Sync.User.adminUser(Token_1.generateAdminToken(this.privateKey), `http://${server.address}`);
        this._syncedAdminRealm = server.getSyncedRealm(realms_1.AdminRealm);
        this._syncedStarPermissionsRealm = server.getSyncedRealm(realms_1.StarPermissionRealm);
        const syncBaseUrl = `realm://${server.address}`;
        const authUrl = `http://${server.address}`;
        this.adminRealm.objects('Permission').sorted('updatedAt').addListener(this.handleAdminRealmPermissionChange);
        this.handleManagementChangeCallback = this.handleManagementRealmChange.bind(this);
        Realm.Sync.addListener(syncBaseUrl, Realm.Sync.User.adminUser(Token_1.generateAdminToken(this.privateKey), authUrl), PermissionService_1.MANAGEMENT_REALM_REGEX, 'change', this.handleManagementChangeCallback);
    }
    stop() {
        Realm.Sync.removeListener(PermissionService_1.MANAGEMENT_REALM_REGEX, 'change', this.handleManagementChangeCallback);
    }
    onDemand(accessToken) {
        if (!accessToken) {
            throw new errors.JSONError({
                title: "X-Realm-Access-Token header is not valid",
                status: 400,
            });
        }
        const token = Token_1.Token.parse(accessToken, this.publicKey);
        const userId = token.identity;
        const permissionsInAdminRealm = this.adminRealm.objects('Permission')
            .filtered('user.userId = $0 || realmFile.owner.userId = $0', userId);
        const userSpecificRealm = this.getUserSpecificPermissionRealm(userId);
        userSpecificRealm.write(() => {
            userSpecificRealm.deleteAll();
            for (let permissionFromAdminRealm of permissionsInAdminRealm) {
                userSpecificRealm.create(realms_1.UserSpecificPermission.schema.name, {
                    mayManage: permissionFromAdminRealm.mayManage,
                    mayRead: permissionFromAdminRealm.mayRead,
                    mayWrite: permissionFromAdminRealm.mayWrite,
                    updatedAt: new Date(),
                    path: permissionFromAdminRealm.realmFile.path,
                    userId: permissionFromAdminRealm.user.userId
                });
            }
        });
        return {
            userId: userId,
            affectedPermissionsCount: permissionsInAdminRealm.length
        };
    }
    getUserByUserId(userId) {
        return this.adminRealm.objectForPrimaryKey('User', userId);
    }
    getAllUsers() {
        return this.adminRealm.objects('User');
    }
    getUserIdsWithMetadata(key, value) {
        let metadata = this.adminRealm.objects('UserMetadataRow');
        if (key) {
            metadata = metadata.filtered("key = $0", key);
        }
        if (value) {
            metadata = metadata.filtered("value = $0", value);
        }
        return Array.from(new Set(metadata.map((meta) => meta.user.userId)));
    }
    getUsersByUserIds(userIds) {
        const query = userIds.map((id, index) => {
            return `userId = $${index}`;
        }).join(" || ");
        return this.adminRealm.objects("User").filtered(query, ...userIds);
    }
    getManagementRealmForUserId(userId) {
        const realmDefinition = realms_1.CreateUserManagementRealmDefinition(userId);
        const syncedRealm = this.server.getSyncedRealm(realmDefinition);
        return syncedRealm.realm;
    }
    getAdminRealmPermissionForUserAndFile(realmFilePath, userId) {
        const realmFile = this.adminRealm.objectForPrimaryKey("RealmFile", realmFilePath);
        const user = this.getUserByUserId(userId);
        if (!realmFile) {
            throw new errors.realm.InvalidParameters('realmFilePath does not exist');
        }
        let permission = realmFile.permissions.filtered('user.userId = $0', userId)[0];
        if (!permission) {
            this.adminRealm.write(() => {
                realmFile.permissions.push({
                    user: user,
                    realmFile: realmFile,
                    updatedAt: new Date(),
                    mayManage: false,
                    mayRead: false,
                    mayWrite: false
                });
                permission = realmFile.permissions.filtered('user.userId = $0', userId)[0];
            });
        }
        return {
            permission: permission,
            adminRealm: this.adminRealm
        };
    }
    getUserSpecificPermissionRealm(userId) {
        const realmDefinition = realms_1.CreateUserPermissionRealmDefinition(userId);
        const syncedUserRealm = this.server.getSyncedRealm(realmDefinition);
        return syncedUserRealm.realm;
    }
    getUserSpecificPermissionAndAssociatedRealm(realmPath, userId) {
        const realmDefinition = realms_1.CreateUserPermissionRealmDefinition(userId);
        const syncedUserRealm = this.server.getSyncedRealm(realmDefinition);
        const userSpecificPermission = syncedUserRealm.realm.objects("Permission").filtered('path = $0', realmPath)[0];
        return {
            userSpecificRealm: syncedUserRealm.realm,
            userSpecificPermission: userSpecificPermission
        };
    }
    handleManagementRealmChange(changeEvent) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                const matches = changeEvent.path.match(PermissionService_1.MANAGEMENT_REALM_REGEX);
                if (!matches) {
                    this.logger.error(`Received changeEvent for non-management realm: ${changeEvent.path}`);
                    return;
                }
                const userId = matches[1];
                const user = this.getUserByUserId(userId);
                const realmPath = changeEvent.path;
                const realm = changeEvent.realm;
                this.logger.debug(`A management realm with path ${changeEvent.path} has changes with userId: ${userId}`);
                if (!user) {
                    this.logger.error(`A change notification with management realm of path ${changeEvent.path} 
                was notified with a userId of "${userId}" but this userId was not found in the system. This realm will now be closed.
                Please ensure that ${userId} exists within the __admin Realm. This permission change attempt will be ignored.`);
                    realm.close();
                    return;
                }
                let unprocessedPermissionOfferObjects = realm.objects("PermissionOffer").filtered('statusCode == null');
                let unprocessedPermissionOfferResponseObjects = realm.objects("PermissionOfferResponse").filtered('statusCode == null');
                let unprocessedPermissionChangeObjects = realm.objects("PermissionChange").filtered('statusCode == null');
                for (let p of unprocessedPermissionOfferObjects) {
                    this.handlePermissionOfferObject(p, userId);
                }
                for (let p of unprocessedPermissionOfferResponseObjects) {
                    this.handlePermissionOfferResponseObject(p);
                }
                for (let p of unprocessedPermissionChangeObjects) {
                    this.handlePermissionChangeObject(p, userId, realm);
                }
            }
            catch (err) {
                this.logger.error(`Error in handleManagementRealmChange: ${err.message}`);
            }
        });
    }
    handlePermissionOfferObject(permissionOffer, userId) {
        if (permissionOffer.realmUrl === '') {
            throw new errors.realm.InvalidParameters({
                name: 'realmUrl',
                reason: 'The realmUrl cannot be empty and must encode a path'
            });
        }
        const realmPath = new URI(permissionOffer.realmUrl).path();
        if (_.includes(realmPath, '__management')) {
            throw new errors.realm.InvalidParameters({
                name: 'realmUrl',
                reason: 'The realmUrl cannot be of a management realm'
            });
        }
        const realmFile = this.adminRealm.objectForPrimaryKey("RealmFile", realmPath);
        if (!realmFile) {
            throw new errors.realm.InvalidParameters({
                name: 'realmUrl',
                reason: 'The realmUrl does not exist'
            });
        }
        if (permissionOffer.expiresAt && moment(permissionOffer.expiresAt).isBefore(moment.now())) {
            throw new errors.realm.ExpiredPermissionOffer({
                title: `The permission offer has expired on ${permissionOffer.expiresAt}. The current time is ${new Date()}.`
            });
        }
        this.adminRealm.write(() => {
            permissionOffer.token = `${userId}:${permissionOffer.id}`;
        });
    }
    handlePermissionOfferResponseObject(permissionOfferResponse) {
        const tokenComponents = permissionOfferResponse.token.split(':');
        if (tokenComponents.length < 2 || tokenComponents[0].length === 0) {
            throw new errors.realm.InvalidParameters('token');
        }
        const offeringUserId = tokenComponents[0];
        const offeringRealm = this.getManagementRealmForUserId(offeringUserId);
        const permissionOffers = offeringRealm.objects("PermissionOffer").filtered('token = %0', permissionOfferResponse.token);
        if (permissionOffers.length === 0) {
            throw new errors.realm.InvalidParameters('token');
        }
        if (permissionOffers.length > 1) {
            throw new errors.realm.AmbiguousPermissionOfferToken({
                detail: `Unexpected count of permission offers (=${permissionOffers.length})`
                    + ` matching the token '${permissionOfferResponse.token}' of the permission request!`,
            });
        }
        const permissionOffer = permissionOffers[0];
        permissionOfferResponse.realmUrl = permissionOffer.realmUrl;
        permissionOfferResponse.token = permissionOffer.token;
        const realmPath = new URI(permissionOfferResponse.realmUrl).path();
        const { permission, adminRealm } = this.getAdminRealmPermissionForUserAndFile(realmPath, offeringUserId);
        adminRealm.write(() => {
            permission.mayWrite = permissionOffer.mayWrite;
            permission.mayRead = permission.mayRead;
            permission.mayManage = permission.mayManage;
        });
    }
    handlePermissionChangeObject(permissionChange, userId, realm) {
        try {
            const realmPath = new URI(permissionChange.realmUrl).path();
            if (permissionChange.realmUrl === '*') {
                this.handleStarRealmUrl(realmPath, permissionChange.mayWrite, permissionChange.mayRead, permissionChange.mayManage);
            }
            else if (permissionChange.userId === '*') {
                this.handleStarUserIdUrl(userId, permissionChange.mayWrite, permissionChange.mayRead, permissionChange.mayManage);
            }
            else {
                let realmFiles;
                let users;
                if (permissionChange.metadataKey) {
                    const userIds = this.getUserIdsWithMetadata(permissionChange.metadataKey, permissionChange.metadataValue);
                    this.logger.debug(`Apply permission change by metadata query 
                        (${permissionChange.metadataKey}=${permissionChange.metadataValue}) 
                        matching the userIds ${userIds}`);
                    users = this.getUsersByUserIds(userIds);
                }
                else {
                    this.logger.debug(`Apply permission change for user ${permissionChange.userId}`);
                    users = this.getUsersByUserIds([permissionChange.userId]);
                }
                if (!realmFiles) {
                    realmFiles = this.adminRealm
                        .objects("RealmFile")
                        .filtered('owner.userId = $0', userId);
                }
                this.setPermissionsForEntireSystem(users, realmFiles, permissionChange.mayWrite, permissionChange.mayRead, permissionChange.mayManage);
            }
            realm.write(() => {
                permissionChange.statusCode = 0;
            });
        }
        catch (err) {
            console.log("Failed to apply permission change", err);
            realm.write(() => {
                permissionChange.statusMessage = "Failed to apply permission change";
                permissionChange.statusCode = -1;
            });
        }
    }
    handleStarRealmUrl(path, mayWrite, mayRead, mayManage) {
        this.starPermissionsRealm.write(() => {
            this.starPermissionsRealm.create(realms_1.StarSpecificPermission.schema.name, {
                path: path,
                mayWrite: mayWrite,
                mayManage: mayManage,
                mayRead: mayRead,
                updatedAt: new Date()
            }, true);
        });
    }
    handleStarUserIdUrl(ownerUserId, mayWrite, mayRead, mayManage) {
        const ownedRealmFiles = this.adminRealm.objects(realms_1.RealmFile.schema.name)
            .filtered('owner.userId = $0', ownerUserId);
        this.starPermissionsRealm.write(() => {
            for (let realmFile of ownedRealmFiles) {
                this.starPermissionsRealm.create(realms_1.StarSpecificPermission.schema.name, {
                    path: realmFile.path,
                    mayWrite: mayWrite,
                    mayManage: mayManage,
                    mayRead: mayRead,
                    updatedAt: new Date()
                }, true);
            }
        });
    }
    setPermissionsForEntireSystem(users, realmFiles, mayWrite, mayRead, mayManage, permissionChange) {
        for (let file of realmFiles) {
            for (let user of users) {
                {
                    const { adminRealm, permission } = this.getAdminRealmPermissionForUserAndFile(file.path, user.userId);
                    adminRealm.write(() => {
                        permission.mayManage = mayManage;
                        permission.mayRead = mayRead;
                        permission.mayWrite = mayWrite;
                        permission.updatedAt = new Date();
                    });
                }
                {
                    const { userSpecificRealm, userSpecificPermission } = this.getUserSpecificPermissionAndAssociatedRealm(file.path, user.userId);
                    userSpecificRealm.write(() => {
                        userSpecificRealm.create("Permission", {
                            userId: user.userId,
                            path: file.path,
                            mayManage: mayManage,
                            mayWrite: mayWrite,
                            mayRead: mayRead,
                            updatedAt: new Date()
                        }, true);
                    });
                }
            }
        }
        if (permissionChange) {
            permissionChange.statusMessage = '';
            permissionChange.statusCode = 0;
        }
    }
};
PermissionService.MANAGEMENT_REALM_REGEX = '^/([^/]+)/__management$';
PermissionService.ADMIN_REALM_REGEX = `/__admin`;
__decorate([
    decorators_1.ServerStarted(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Server_1.Server]),
    __metadata("design:returntype", void 0)
], PermissionService.prototype, "serverStarted", null);
__decorate([
    decorators_1.Stop(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", void 0)
], PermissionService.prototype, "stop", null);
__decorate([
    decorators_1.Get('/ondemand'),
    __param(0, decorators_1.Headers('X-Realm-Access-Token')),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [String]),
    __metadata("design:returntype", void 0)
], PermissionService.prototype, "onDemand", null);
PermissionService = PermissionService_1 = __decorate([
    decorators_1.BaseRoute('/permissions'),
    decorators_1.ServiceName('permissions')
], PermissionService);
exports.PermissionService = PermissionService;
var PermissionService_1;
//# sourceMappingURL=PermissionsService.js.map